#include <amsmf/ParseSMF.h>
#include <vsqdump/TextQueue.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

using namespace AMT;
using namespace AMSMF;

namespace Ex
{

using namespace AMSMF::Ex;

struct IllegalSMFFormat : public DataError { };
struct TooFewTracks : public DataError { };
struct DivisionNotSupported : public DataError { };
struct FrameLacked : public DataError { };
struct NoText : public DataError { };

}

class FileStream : public Parsers::InputStream
{
	FILE *m_fp;
public:
	FileStream(const char *file) : m_fp(fopen(file, "rb"))
	{
		if (!m_fp)
			throw AMT::Ex::CannotOpenFile();
	}
	~FileStream()
	{
		if (m_fp)
			fclose(m_fp);
	}
	Result read(void *ptr, U32 *rlen)
	{
		size_t len;

		len = fread(ptr, 1, *rlen, m_fp);
		*rlen = (U32)len;
		if (len == 0) {
			if (ferror(m_fp))
				throw AMT::Ex::CannotReadFile();
			return R_EOF;
		}
		return R_SUCCESS;
	}
	Result peek(U8 *byte)
	{
		int ch = getc(m_fp);
		if (ch == EOF) {
			if (ferror(m_fp))
				throw AMT::Ex::CannotReadFile();
			return R_EOF;
		}
		*byte = (U8)ch;
		ungetc(ch, m_fp);
		return R_SUCCESS;
	}
};

const int VSQ_SMF_FORMAT = 1;
const int VSQ_MIN_NUM_TRACKS = 2;
const int VSQ_MASTER_TRACK_NO = 0;
const char VSQ_TEXT_ID[] = "DM:";
const Size VSQ_TEXT_ID_LEN = sizeof (VSQ_TEXT_ID) -1;
const char VSQ_TEXT_ID_SEP = ':';

class VSQParserHooks : public Parsers::Hooks
{
	static const int INITIAL_DM_NUM_ = 0;

	FILE *m_out;
	bool m_out_need_close;
	int m_num_tracks;
	int m_cur_dm_num;
	int m_division;
	U32 m_curdelta;
	TextQueue m_queue;
public:
	VSQParserHooks() : m_out(stdout), m_out_need_close(false) { }
	~VSQParserHooks()
	{
		if (m_out_need_close)
			fclose(m_out);
	}
	void begin_smf(const SMFHeader &hdr)
	{
		if (hdr.format() != VSQ_SMF_FORMAT)
			throw ::Ex::IllegalSMFFormat();
		m_num_tracks = hdr.num_tracks();
		if (m_num_tracks < VSQ_MIN_NUM_TRACKS)
			throw ::Ex::TooFewTracks();
		if (!DivClock::is_match(hdr))
			throw ::Ex::DivisionNotSupported();
		m_division = DivClock(hdr).num_clocks();
	}
	void end_smf()
	{
	}
	void begin_header_extra_bytes(U32) { }
	void end_header_extra_bytes() { }
	void header_extra_bytes(const U8 *, U32) { }
	void begin_track(int tr, U32)
	{
		if (tr == VSQ_MASTER_TRACK_NO)
			this->begin_master_track_();
		else
			this->begin_voice_track_(tr);
		m_curdelta = 0;
	}
	void end_track(int tr)
	{
		if (tr != VSQ_MASTER_TRACK_NO) {
			if (!m_queue.is_empty())
				throw ::Ex::FrameLacked();
			if (m_cur_dm_num == INITIAL_DM_NUM_-1)
				throw ::Ex::NoText();
		}
		if (tr != m_num_tracks-1)
			chkprf_(fprintf(m_out, "\n"));
	}
	void track_event(int tr, const TrackEvent &ev)
	{
		m_curdelta += ev.delta();
		if (tr == VSQ_MASTER_TRACK_NO)
			this->process_master_track_(ev);
		else
			this->process_voice_track_(tr, ev);
	}
private:
	static void chkprf_(int ret)
	{
		if (ret < 0)
			throw ::Ex::CannotWriteFile();
	}
	static const char *check_vsq_text_(const char *text, int *rnum)
	{
		unsigned long num;
		char *endptr;

		if (text != strstr(text, VSQ_TEXT_ID))
			return NULL;
		text += VSQ_TEXT_ID_LEN;
		num = strtoul(text, &endptr, 10);
		if (*endptr != VSQ_TEXT_ID_SEP)
			return NULL;
		*rnum = (int)num;
		return endptr+1;
	}
	void begin_master_track_()
	{
		chkprf_(fprintf(m_out, "[Master]\n"));
		chkprf_(fprintf(m_out, "NumVoices=%d\n", m_num_tracks-1));
		chkprf_(fprintf(m_out, "Division=%d\n", m_division));
		chkprf_(fprintf(m_out, "[TempoMap]\n"));
	}
	void process_master_track_(const TrackEvent &ev)
	{
		if (Parsers::SMFMetaSetTempo::is_match(ev)) {
			Parsers::SMFMetaSetTempo rd(ev);
			chkprf_(fprintf(m_out, "%u=Tempo:%.2f\n",
					(unsigned)m_curdelta,
					60000000.0/(double)rd.tempo()));
		} else if (Parsers::SMFMetaTimeSign::is_match(ev)) {
			Parsers::SMFMetaTimeSign rd(ev);
			chkprf_(fprintf(m_out, "%u=Time:%u/%u:%u:%u\n",
					(unsigned)m_curdelta,
					(unsigned)rd.numerator(),
					(unsigned)rd.calc_denominator(),
					(unsigned)rd.num_clocks_per_click(),
					(unsigned)
					rd.num_32nd_notes_per_quater()));
		}
	}
	void begin_voice_track_(int tr)
	{
		m_cur_dm_num = INITIAL_DM_NUM_-1;
		chkprf_(fprintf(m_out, "[Track%02d]\n", tr));
	}
	void process_voice_track_(int, const TrackEvent &ev)
	{
		int num;
		const char *text;

		if (Parsers::SMFMetaText::is_match(ev)) {
			Parsers::SMFMetaText rd(ev);
			text = check_vsq_text_(rd.text(), &num);
			if (text) {
				if (num != m_cur_dm_num+1) {
					// confusing the DM order
					m_queue.push(num, text);
				} else {
					chkprf_(fprintf(m_out, "%s", text));
					m_cur_dm_num++;
				}
				while (!m_queue.is_empty() &&
				       m_queue.top_num()==m_cur_dm_num+1) {
					chkprf_(fprintf(m_out, "%s",
							m_queue.top_text()));
					m_cur_dm_num++;
					m_queue.pop();
				}
			}
		}
	}
};

void
usage()
{
	fprintf(stderr, "usage: vsqdump <file>\n");
	exit(EXIT_FAILURE);
}

void
err(const char *msg)
{
	fprintf(stderr, "Error: %s", msg);
	exit(EXIT_FAILURE);
}

int
main(int argc, char **argv)
{
	if (argc != 2)
		usage();

	try {
		FileStream s(argv[1]);
		VSQParserHooks h;
		Parser p(s, h);
		while (p.parse())
			;
	}
	catch (::Ex::IllegalSMFFormat &) {
		err("Illegal SMF Format (Format 1 required).\n");
	}
	catch (::Ex::TooFewTracks &) {
		err("Too few number of tracks.\n");
	}
	catch (::Ex::DivisionNotSupported &) {
		err("SMPTE style division is not supported.\n");
	}
	catch (::Ex::FrameLacked &) {
		err("Some DM frames lacked.\n");
	}
	catch (::Ex::NoText &) {
		err("No DM frames - SMF, but not VSQ.\n");
	}
	catch (::Ex::UnexpectedFileEnd &) {
		err("Unexpected file end.\n");
	}
	catch (::Ex::IllegalMagic &) {
		err("Illegal SMF magic.\n");
	}
	catch (::Ex::ChunkOverrun &) {
		err("SMF data error - track chunk overrun.\n");
	}
	catch (::Ex::IllegalRunningStatus &) {
		err("SMF data error - illegal running status.\n");
	}
	catch (::Ex::TooShortHeader &) {
		err("Too short SMF header.\n");
	}
	catch (::Ex::IllegalSystemMessage &) {
		err("SMF data error - "
			"raw system message is disallowed.\n");
	}
	catch (::Ex::IllegalStatusByte &) {
		err("SMF data error - illegal status byte.\n");
	}
	catch (::Ex::TooLongVarInteger &) {
		err("SMF data error - "
			"variable length integer overflow.\n");
	}
	catch (::Ex::CannotOpenFile &) {
		err("cannot open file.\n");
	}
	catch (::Ex::CannotReadFile &) {
		err("cannot read file.\n");
	}
	catch (::Ex::CannotWriteFile &) {
		err("cannot write file.\n");
	}
	catch (...) {
		err("exception caught.\n");
	}

	return 0;
}
